<?php

namespace Unific\Connector\Helper;

use Laminas\Http\Request;
use Magento\Catalog\Model\ResourceModel\Category\Collection as CategoryCollection;
use Magento\Catalog\Model\ResourceModel\Category\CollectionFactory as CategoryCollectionFactory;
use Magento\Catalog\Model\ResourceModel\Product\Collection as ProductCollection;
use Magento\Catalog\Model\ResourceModel\Product\CollectionFactory as ProductCollectionFactory;
use Magento\Customer\Model\ResourceModel\CustomerRepository;
use Magento\Framework\Api\SearchCriteriaBuilder;
use Magento\Framework\App\Helper\AbstractHelper;
use Magento\Framework\App\Helper\Context;
use Magento\Framework\Exception\NoSuchEntityException;
use Magento\Sales\Model\OrderRepository;
use Magento\Store\Api\Data\StoreInterface;
use Magento\Store\Model\App\Emulation;
use Magento\Store\Model\StoreManagerInterface;
use Unific\Connector\Helper\Data\Cart;
use Unific\Connector\Helper\Data\Category;
use Unific\Connector\Helper\Data\Customer;
use Unific\Connector\Helper\Data\Order;
use Unific\Connector\Helper\Data\Product;
use Unific\Connector\Helper\Message\Queue;
use Unific\Connector\Model\HistoricalFactory;
use Unific\Connector\Model\ResourceModel\Historical\Collection as HistoricalCollection;
use Unific\Connector\Model\ResourceModel\Historical\CollectionFactory as HistoricalCollectionFactory;
use Unific\Connector\Model\ResourceModel\Queue\Collection as QueueCollection;
use Unific\Connector\Model\ResourceModel\Queue\CollectionFactory as QueueCollectionFactory;

class Historical extends AbstractHelper
{
    /**
     * @var Message\Queue
     */
    protected $queueHelper;
    /**
     * @var QueueCollectionFactory
     */
    protected $queueCollectionFactory;
    /**
     * @var HistoricalCollectionFactory
     */
    protected $historicalCollectionFactory;
    /**
     * @var HistoricalFactory
     */
    protected $historicalFactory;
    /**
     * @var SearchCriteriaBuilder
     */
    protected $searchCriteriaBuilder;
    /**
     * @var OrderRepository
     */
    protected $orderRepository;
    /**
     * @var CustomerRepository
     */
    protected $customerRepository;
    /**
     * @var CategoryCollectionFactory
     */
    protected $categoryCollectionFactory;
    /**
     * @var ProductCollectionFactory
     */
    protected $productCollectionFactory;
    /**
     * @var Data\Order
     */
    protected $orderDataHelper;
    /**
     * @var Data\Customer
     */
    protected $customerDataHelper;
    /**
     * @var Data\Cart
     */
    protected $cartDataHelper;
    /**
     * @var Data\Product
     */
    protected $productDataHelper;
    /**
     * @var Data\Category
     */
    protected $categoryDataHelper;
    /**
     * @var Emulation
     */
    protected $emulation;
    /**
     * @var StoreManagerInterface
     */
    protected $storeManager;

    /**
     * @param Context $context
     * @param QueueCollectionFactory $queueCollectionFactory
     * @param HistoricalCollectionFactory $historicalCollectionFactory
     * @param HistoricalFactory $historicalFactory
     * @param Queue $queueHelper
     * @param SearchCriteriaBuilder $searchCriteriaBuilder
     * @param OrderRepository $orderRepository
     * @param CustomerRepository $customerRepository
     * @param ProductCollectionFactory $productCollectionFactory
     * @param CategoryCollectionFactory $categoryCollectionFactory
     * @param Order $orderDataHelper
     * @param Cart $cartDataHelper
     * @param Customer $customerDataHelper
     * @param Product $productDataHelper
     * @param Category $categoryDataHelper
     * @param Emulation $emulation
     * @param StoreManagerInterface $storeManager
     */
    public function __construct(
        Context                     $context,
        QueueCollectionFactory      $queueCollectionFactory,
        HistoricalCollectionFactory $historicalCollectionFactory,
        HistoricalFactory           $historicalFactory,
        Queue                       $queueHelper,
        SearchCriteriaBuilder       $searchCriteriaBuilder,
        OrderRepository             $orderRepository,
        CustomerRepository          $customerRepository,
        ProductCollectionFactory    $productCollectionFactory,
        CategoryCollectionFactory   $categoryCollectionFactory,
        Order                       $orderDataHelper,
        Cart                        $cartDataHelper,
        Customer                    $customerDataHelper,
        Product                     $productDataHelper,
        Category                    $categoryDataHelper,
        Emulation                   $emulation,
        StoreManagerInterface       $storeManager
    ) {
        parent::__construct($context);

        $this->queueHelper = $queueHelper;
        $this->searchCriteriaBuilder = $searchCriteriaBuilder;
        $this->orderRepository = $orderRepository;
        $this->customerRepository = $customerRepository;
        $this->productCollectionFactory = $productCollectionFactory;
        $this->categoryCollectionFactory = $categoryCollectionFactory;
        $this->queueCollectionFactory = $queueCollectionFactory;
        $this->historicalCollectionFactory = $historicalCollectionFactory;
        $this->historicalFactory = $historicalFactory;
        $this->orderDataHelper = $orderDataHelper;
        $this->customerDataHelper = $customerDataHelper;
        $this->cartDataHelper = $cartDataHelper;
        $this->productDataHelper = $productDataHelper;
        $this->categoryDataHelper = $categoryDataHelper;
        $this->emulation = $emulation;
        $this->storeManager = $storeManager;
    }

    /**
     * Trigger the historical process
     *
     * @return array
     * @throws \Exception
     */
    public function triggerHistorical()
    {
        $this->resetAllHistoricalData();

        $this->triggerHistoricalForType('customer');
        $this->triggerHistoricalForType('order');
        $this->triggerHistoricalForType('category');
        $this->triggerHistoricalForType('product');

        return ['message' => 'Historical data has been triggered'];
    }

    /**
     * Trigger historical for a specific type
     *
     * @param $type
     * @param bool $removeOld
     *
     * @return array
     * @throws \Exception
     */
    public function triggerHistoricalForType($type, $removeOld = false)
    {
        if ($removeOld) {
            $this->removeHistoricalQueueType($type);
        }

        try {
            /** @var \Unific\Connector\Model\Historical $historical */
            $historical = $this->historicalFactory->create();
            $historical->setHistoricalType($type);
            $historical->setHistoricalTypePage(1);
            $historical->save();
        } catch (\Exception $e) {
            $this->_logger->critical($e);
        }

        return ['message' => 'Historical data for type [' . $type . '] has been triggered'];
    }

    /**
     * Delete a type from the historical queue process
     *
     * @param $type
     */
    public function removeHistoricalQueueType($type)
    {
        /** @var HistoricalCollection $collection */
        $collection = $this->historicalCollectionFactory->create();
        $collection->addFieldToFilter('historical_type', ['eq' => $type]);
        $collection->walk('delete');
    }

    /**
     * Inject historical data in the queue if historical is requested
     */
    public function queueHistorical()
    {
        /** @var HistoricalCollection $historicalCollection */
        $historicalCollection = $this->historicalCollectionFactory->create();

        if ($historicalCollection->getSize() === 0) {
            return;
        }

        foreach ($historicalCollection as $historicalItem) {
            try {
                $items = $this->processHistoricalItem($historicalItem);
                // Write back the latest ID to the historical queue manager
                if (count($items) == Settings::HISTORICAL_PAGE_SIZE) {
                    $this->incrementHistoricalQueuePage($historicalItem->getHistoricalType());
                } else {
                    $this->removeHistoricalQueueType($historicalItem->getHistoricalType());
                }
            } catch (\Exception $e) {
                $this->_logger->critical($e);
            }
        }
    }

    /**
     * Queue up all the historical data to be ready for sending
     */
    public function resetAllHistoricalData()
    {
        /** @var HistoricalCollection $collection */
        $collection = $this->historicalCollectionFactory->create();
        $collection->walk('delete');

        $this->resetHistoricalForType();

        return ['message' => 'Historical data has been stopped and reset'];
    }

    /**
     * Queue up all the historical data to be ready for sending
     * @param $type
     * @return array
     */
    public function resetHistoricalForType($type = null)
    {
        if ($type) {
            $this->removeHistoricalQueueType($type);
        }

        /** @var QueueCollection $queueCollection */
        $queueCollection = $this->queueCollectionFactory->create();
        $queueCollection->addFieldToFilter('headers', ['like' => '%' . $type . '/historical%']);

        $queueCollection->walk('delete');

        return ['message' => 'Historical data for type [' . $type . '] has been stopped and reset'];
    }

    /**
     * Update the latest queued ID
     *
     * @param $type
     */
    protected function incrementHistoricalQueuePage($type)
    {
        /** @var HistoricalCollection $collection */
        $collection = $this->historicalCollectionFactory->create();
        $collection->addFieldToFilter('historical_type', ['eq', $type]);

        foreach ($collection as $item) {
            $item->setHistoricalTypePage($item->getHistoricalTypePage() + 1);
        }

        $collection->save();
    }

    /**
     * @param $subject
     * @return array
     */
    protected function getHeaders($subject)
    {
        $headers = [];
        $headers['X-SUBJECT'] = $subject;

        return $headers;
    }

    /**
     * @param \Unific\Connector\Model\Historical $historicalItem
     * @return mixed
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    protected function processHistoricalItem(\Unific\Connector\Model\Historical $historicalItem)
    {
        $items = $this->getTypeData($historicalItem, 'collection');
        [$type] = explode(
            false !== strpos((string)$historicalItem->getHistoricalType(), ':') ? ':' : '_',
            (string)$historicalItem->getHistoricalType()
        );

        $this->createQueueEntries(
            $items,
            $this->getTypeData($historicalItem, 'helper'),
            $type,
            $this->getTypeData($historicalItem, 'priority')
        );

        return $items;
    }

    /**
     * @param iterable $items
     * @param $helper
     * @param string $type
     * @param int $priority
     */
    protected function createQueueEntries($items, $helper, $type, $priority)
    {
        $setEntityMethod = 'set' . ucwords($type);
        $getInfoMethod = 'get' . ucwords($type) . 'Info';
        foreach ($items as $item) {
            $helper->$setEntityMethod($item);
            if ($type === 'order') {
                if ($item->hasShipments()) {
                    $helper->setShipment($item->getShipmentsCollection()->getLastItem());
                }
            }
            if ((int)$item->getStoreId() > 0) {
                try {
                    $this->emulation->startEnvironmentEmulation($item->getStoreId(), 'frontend');
                } catch (\Throwable $e) {
                    $this->_logger->notice('Unable to emulate store ' . $item->getStoreId());
                }
            }

            $headers = $this->getHeaders($type . '/historical');

            $this->queueHelper->queue(
                $this->scopeConfig->getValue('unific/webhook/' . $type . '_endpoint'),
                $helper->$getInfoMethod(),
                $priority,
                $headers,
                Request::METHOD_POST,
                true,
                null,
                null,
                Settings::QUEUE_HISTORICAL_MAX_RETRIES
            );

            $this->emulation->stopEnvironmentEmulation();
        }
    }

    /**
     * @param \Unific\Connector\Model\Historical $historical
     * @param string $element
     * @return mixed
     * @throws \Magento\Framework\Exception\LocalizedException
     * @todo if historical is set to save to file maybe increase Settings::HISTORICAL_PAGE_SIZE to make the job require less runs
     */
    protected function getTypeData(\Unific\Connector\Model\Historical $historical, string $element)
    {
        $helper = $element === 'helper';
        $type = (string)$historical->getHistoricalType();
        if (false !== strpos($historical->getHistoricalType(), ':')) {
            $elements = explode(':', (string)$historical->getHistoricalType());
            $type = $elements[0];
            unset($elements[0]);
            $store = count($elements) > 1 ? implode(':', $elements) : reset($elements);
        } elseif (false !== strpos($historical->getHistoricalType(), '_')) {
            [$type, $id] = array_pad(explode('_', (string)$historical->getHistoricalType()), 2, null);
        }
        switch ($type) {
            case 'customer':
                if ($element === 'collection') {
                    $this->searchCriteriaBuilder->setPageSize(Settings::HISTORICAL_PAGE_SIZE)
                        ->setCurrentPage($historical->getHistoricalTypePage());
                    if (isset($store)) {
                        $this->searchCriteriaBuilder->addFilter('store_id', ['eq' => $this->getStoreId($store)]);
                    }
                    if (isset($id)) {
                        $this->searchCriteriaBuilder->addFilter('entity_id', ['eq' => $id]);
                    }
                    $searchCriteria = $this->searchCriteriaBuilder->create();

                    return $this->customerRepository->getList($searchCriteria)->getItems();
                }
                return $helper ? $this->customerDataHelper : Settings::PRIORITY_CUSTOMER;
            case 'order':
                if ($element === 'collection') {
                    $this->searchCriteriaBuilder->setPageSize(Settings::HISTORICAL_PAGE_SIZE)
                        ->setCurrentPage($historical->getHistoricalTypePage());
                    if (isset($store)) {
                        $this->searchCriteriaBuilder->addFilter('store_id', ['eq' => $this->getStoreId($store)]);
                    }
                    if (isset($id)) {
                        $this->searchCriteriaBuilder->addFilter('entity_id', ['eq' => $id]);
                    }
                    $searchCriteria = $this->searchCriteriaBuilder->create();

                    return $this->orderRepository->getList($searchCriteria)->getItems();
                }
                return $helper ? $this->orderDataHelper : Settings::PRIORITY_ORDER;
            case 'product':
                if ($element === 'collection') {
                    /** @var ProductCollection $items */
                    $items = $this->productCollectionFactory->create();
                    $items->addAttributeToSelect('*')
                        ->setPageSize(Settings::HISTORICAL_PAGE_SIZE)
                        ->setCurPage($historical->getHistoricalTypePage());
                    if (isset($store) && $storeId = $this->getStoreId($store)) {
                        $items->addStoreFilter($storeId);
                    }
                    if (isset($id)) {
                        $items->addIdFilter($id);
                    }
                    $items->load()
                        ->addCategoryIds();

                    return $items;
                }
                return $helper ? $this->productDataHelper : Settings::PRIORITY_PRODUCT;
            case 'category':
                if ($element === 'collection') {
                    /** @var CategoryCollection $items */
                    $items = $this->categoryCollectionFactory->create();
                    $items->addAttributeToSelect('*')
                        ->setPageSize(Settings::HISTORICAL_PAGE_SIZE)
                        ->setCurPage($historical->getHistoricalTypePage());
                    if (isset($store)) {
                        $items->setStoreId($this->getStoreId($store));
                    }
                    if (isset($id)) {
                        $items->addIdFilter($id);
                    }

                    return $items;
                }
                return $helper ? $this->categoryDataHelper : Settings::PRIORITY_CATEGORY;
        }

        throw new \InvalidArgumentException('Unknown historical type: ' . $historical->getHistoricalType());
    }

    protected function getStoreId($storeCode): ?int
    {
        try {
            $store = $this->storeManager->getStore($storeCode);
            return $store->getId();
        } catch (NoSuchEntityException $e) {
            $this->_logger->warning('Cannot find store with code ' . $storeCode);
        }

        return null;
    }
}
